#include <Servo.h>
#include <Adafruit_NeoPixel.h> //Download from Adafruit's Website
#define PIN 1
/* Detects patterns of knocks and triggers a servo to dispatch a gumball
 
 By Steve Hoefer http://grathio.com
 Version 10.12.20
 Licensed under Creative Commons Attribution-Noncommercial-Share Alike 3.0
 http://creativecommons.org/licenses/by-nc-sa/3.0/us/
 (In short: Do what you want, as long as you credit me, don't relicense it and don't sell it or use it in anything you sell without contacting me.)
 
 Analog Pin 0: Piezo speaker (connected to ground with 480K ohm pulldown resistor)
 Analog Pin 1: 10K Trim pot.
 Digital Pin 2: Switch to enter new knock code.
 Digital Pin 3: Success LED. 
 Digital Pin 4: Failure LED. 
 Digital Pin 5: Activity light 1. (Optional: lights up when the servo is turning.) 
 Digital Pin 6: Activity light 2. (Optional: lights up the gumball end point)
 Digital Pin 7: Servo motor signal.
 
 Version 10.12.20:
 - Commented out the Serial output to significantly increase senitvity.
 - Improved debounce code so rapid knocks are hear better.
 - Tray light no longer stays on for  many minutes.
 - Slightly increased sensitivity through general code changes.
 - Cleaned up comments and formatting.
 
 Version 10.11.11: Initial release
 
 */
// Pin definitions
const int knockSensor = A0;         // Piezo sensor on analog pin 0.
const int trimPot = A3;             // Trim potentiometer that allows the user to set the sensitivity
const int programSwitch = 3;       // When this is high we program a new code
const int servoPin = 7;            // Servo that we're controlling. 
const int redLED = 4;              // Status LED for failed knock
//const int greenLED = 1;            // Status LED for successful knock.
const int programmled = 5;        // Lights up the servo when its working. (Optional)
const int activityLED2 = 6;        // Gumball tray light. (Optional)
// Tuning constants.  Could be made vars and hooked to potentiometers for soft configuration, etc.
const int thresholdMax = 40;       // Maximum value we can set for the knock sensitivity based on the trim pot.  Increase this if it's too sensitive.
const int rejectValue = 30;        // If an individual knock is off by this percentage of a knock we ignore. (30 is pretty lose. 10 is strict)
const int averageRejectValue = 20; // If the average timing of the knocks is off by this percent we ignore. (20 is pretty lose, 10 is strict.)
const int debounceThreshold = 80;  // Simple debounce timer to make sure we don't register more than one knock.
const int maximumKnocks = 20;       // Maximum number of knocks to listen for.
const int knockComplete = 1500;     // Longest time to wait for a knock before we assume that it's finished.
// Variables.
Servo gumballServo;                  // The servo that dispenses a gumball
int secretCode[maximumKnocks] = {
  50, 25, 25, 50, 100, 50, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};  // Initial setup: "Shave and a Hair Cut, two bits." 100=full note, 50=half note, 25=quarter note, etc.
int knockReadings[maximumKnocks];    // When someone knocks this array fills with delays between knocks. (A correct knock looks a lot like the line above).
int knockSensorValue = 0;            // Most recent reading of the knock sensor.
int threshold =1023;                 // Minimum signal from the piezo to register as a knock.  (Set by the trim pot.)
int programButtonPressed = false;    // Flag so we remember the programming button setting at the end of the cycle.
long counter = 0;                    // Used to time the display light and the trim pot readings.
int buttonState = LOW;             // the current reading from the input pin
Adafruit_NeoPixel strip = Adafruit_NeoPixel(17, PIN, NEO_GRB + NEO_KHZ800); //Sets up the RGB strip. Change first number to number of pixels on strip.

void setup() {
  strip.begin();
  strip.show(); // Initialize all pixels to 'off'
  pinMode(programSwitch, INPUT);
  pinMode(programmled, OUTPUT);
  threshold = map(analogRead(trimPot), 0, 1023, 0, thresholdMax);  // Initial reading of the trim pot.

  gumballServo.attach(servoPin);   // Initialize the servo.
  gumballServo.write(25);
  delay(50);
  gumballServo.detach();           // Detach the servo when not using it.

  //Serial.begin(9600);               // To debug search and replace (Ctl-F) "////Serial." with "////Serial." 
  //Serial.println("Program start."); // NOTE: This will reduce the sensitivity of the knock sensor.

}
void loop() {
  knockSensorValue = analogRead(knockSensor);
  int reading = digitalRead(programSwitch);
  buttonState = reading;
  
  counter++;
  threshold = map(analogRead(trimPot), 0, 1023, 0, thresholdMax);  // Read the trim pot to see what our lower limit is.
      
      if (buttonState == HIGH) {
        //Serial.println("HIGH");
        programButtonPressed = true;
        
        //delay(1000);
      } else {
        //Serial.println("LOW");
        programButtonPressed = false;
        //delay(1000);
        
      }
  if (knockSensorValue > threshold){
    listenToSecretKnock();
  }
} 

// Records the timing of knocks.
void listenToSecretKnock(){
  //Serial.println("(knock started)");   
  pinMode(programSwitch, INPUT);
  //digitalWrite(programmled , LOW); // Turn off any other lights that might be on.
  int i = 0;
  int counter = 0;
  // First lets reset the listening array.
  for (i=0;i<maximumKnocks;i++){
    knockReadings[i]=0;
  }

  int currentKnockNumber=0;         	         // Increment for the array.
  int startTime=millis();           	         // Reference for when this knock started.
  int now=millis();

  digitalWrite(programmled , HIGH);     	        // Light the green LED for knock feedback.

  delay(debounceThreshold);
  digitalWrite(programmled , LOW);

  do {
    // Listen for the next knock or wait for it to time out. 
    knockSensorValue = analogRead(knockSensor);
    threshold = map(analogRead(trimPot), 0, 1023, 0, thresholdMax);  

    if (knockSensorValue > threshold){                   // Got another knock...
      //Serial.println(" knock!");
      now=millis();                                      // Record the delay time.
      knockReadings[currentKnockNumber] = now-startTime;
      currentKnockNumber ++;                             // Increment the counter.
      startTime=now;          
      digitalWrite(programmled , HIGH);       	         // Light the knock LED.

      delay(debounceThreshold/2);                        // Debounce the knock sensor. (And turn the LED off halfway through.)
      digitalWrite(programmled , LOW);   
      delay(debounceThreshold/2);   
    }

    now=millis();

    // Did we timeout or have too many knocks?
  } 
  while ((now-startTime < knockComplete) && (currentKnockNumber < maximumKnocks));

  // We have a completed knock, lets see if it's valid
  if (programButtonPressed==false){  
    if (validateKnock() == true){      
      triggerSuccessfulAction(); 
    } 
    else {
      triggerFailedAction(); 
    }
  } 
  else { // If we're in programming mode we still run it through the validate function in order to get some useful numbers.
    validateKnock();
    // Blink the green and red lights in sequence to show that program is complete.
    //Serial.println("New lock stored.");
    
  }
}

// We got a good knock, so do something!
void triggerSuccessfulAction(){
  theaterChase(strip.Color(0, 255, 255), 1000); // Blue
  gumballServo.attach(servoPin);      // Now we start with the servo.
  //Serial.println("Knock Code Perfect!");
  ////start////
gumballServo.write(25); //END
delay(250);
gumballServo.write(110); //START

delay(2000);
gumballServo.write(20);
delay(100);
gumballServo.write(160);
delay(1000);
gumballServo.write(20);
delay(100);
gumballServo.write(175);
delay(100);
gumballServo.write(90);
////end////
  
  delay(150);    
  gumballServo.detach();             // Done with the servo, so detach it.
  colorWipe(strip.Color(0, 0, 0), 0); // Turns the RGB Strip Off
  strip.show();
  delay(100);
}


// We didn't like the knock.  Indicate displeasure.
void triggerFailedAction(){
  //Serial.println("Secret knock failed.");
  for (int i=0;i<4;i++){					
    colorWipe(strip.Color(255, 0, 0), 50); // Red
    delay(100);
    strip.show();
    colorWipe(strip.Color(0, 0, 0), 0); // Red
    delay(100);
  }
}

// Checks if our knock matches the secret.
// Returns true if it's a good knock, false if it's not.
boolean validateKnock(){
  int i=0;

  // Simplest check first: Did we get the right number of knocks?
  int currentKnockCount = 0;
  int secretKnockCount = 0;
  int maxKnockInterval = 0;             // We use this later to normalize the times.

  for (i=0;i<maximumKnocks;i++){
    if (knockReadings[i] > 0){
      currentKnockCount++;
    }
    if (secretCode[i] > 0){  					
      secretKnockCount++;
    }

    if (knockReadings[i] > maxKnockInterval){ 	// Collect normalization data while we're looping.
      maxKnockInterval = knockReadings[i];
    }
  }

  // If we're recording a new knock, save the relevant info and get out of here.
  if (programButtonPressed==true){
    for (i=0;i<maximumKnocks;i++){ // Normalize the knock timing
      secretCode[i]= map(knockReadings[i],0, maxKnockInterval, 0, 100);
    }
    // And flash the lights in the recorded pattern to let us know it's been programmed.
    delay(750);

    //Start playing back the knocks
    digitalWrite(redLED, HIGH);  // First knock
    delay(40);
    for (i = 0; i < maximumKnocks ; i++){
      digitalWrite(redLED, LOW);  

      if (programButtonPressed==true){  // Only turn it on if there's a delay
        if (secretCode[i] > 0){                                   
          delay(map(secretCode[i],0, 100, 0, maxKnockInterval)); // Expand the time back out to what it was.  Roughly. 
          digitalWrite(programmled, HIGH);
        }
      }
      delay(40);
      digitalWrite(programmled, LOW); 
    }
    return false; 	// We don't do anything when we are recording a new knock.
  }

  if (currentKnockCount != secretKnockCount){
    return false;   // Return false if the number of knocks are wrong.
  }

  /*  Now we compare the relative intervals of our knocks, not the absolute time between them.
   (ie: if you do the same pattern slow or fast it should still work.)
   This makes it less picky, which does make it less secure but also makes it
   less of a pain to use if you're tempo is a little slow or fast. 
   */
  boolean codeFound=true;
  int totaltimeDifferences=0;
  int timeDiff=0;

  for (i=0;i<maximumKnocks;i++){ // Normalize the times
    knockReadings[i]= map(knockReadings[i],0, maxKnockInterval, 0, 100);      
    timeDiff = abs(knockReadings[i]-secretCode[i]);
    if (timeDiff > rejectValue){ // Individual value too far out of whack
      codeFound=false;
    }
    totaltimeDifferences += timeDiff;
  }
  // It can also fail if the whole thing is too inaccurate.
  if (totaltimeDifferences/secretKnockCount>averageRejectValue){
    codeFound = false;
  }

  if (codeFound==false){
    return false;
  } 
  else {
    return true;
  }

}
void colorWipe(uint32_t c, uint8_t wait) {
  for(uint16_t i=0; i<strip.numPixels(); i++) {
      strip.setPixelColor(i, c);
      strip.show();
      delay(wait);
  }
}

void theaterChase(uint32_t c, uint8_t wait) {
  for (int j=0; j<10; j++) {  //do 10 cycles of chasing
    for (int q=0; q < 3; q++) {
      for (int i=0; i < strip.numPixels(); i=i+3) {
        strip.setPixelColor(i+q, c);    //turn every third pixel on
      }
      strip.show();
     
      delay(wait);
     
      for (int i=0; i < strip.numPixels(); i=i+3) {
        strip.setPixelColor(i+q, 0);        //turn every third pixel off
      }
    }
  }
}

uint32_t Wheel(byte WheelPos) {
  if(WheelPos < 85) {
   return strip.Color(WheelPos * 3, 255 - WheelPos * 3, 0);
  } else if(WheelPos < 170) {
   WheelPos -= 85;
   return strip.Color(255 - WheelPos * 3, 0, WheelPos * 3);
  } else {
   WheelPos -= 170;
   return strip.Color(0, WheelPos * 3, 255 - WheelPos * 3);
  }
}
